GSR_ROUTER_NAME = "GSR_ROUTER_COMPONENT"
GSR_COORDINATOR_NAME = "GSR_COORDINATOR_COMPONENT"
GSR_APPLICATION_NAME = "GSR_APPLICATION_COMPONENT"
GSR_TESTING_NAME = "GSR_TESTING_COMPONENT"

TERMINATE_MESSAGE_TYPE = "GSRTERMINATE"
INFO_MESSAGE_TYPE = "INFOMESSAGE"
UPDATE_MESSAGE_TYPE = "GSRUPDATE"
TERMINATE_ROUTING_MESSAGE_TYPE = "GSRTERMINATEROUTING"
ROUTING_COMPLETED_MESSAGE_TYPE = "GSRROUTINGCOMPLETED"

N_NODES = 10
ROUTER_UPDATE_PERIOD_SECS = 0.1
TOMBALA_DRAW_PERIOD_SECS = 0.2
TOMBALA_PLAYER_PERIOD_SECS = 0.1
TOMBALA_MAX_NUMBER = 90
TOMBALA_N_NUMBERS_IN_HAND = 9

ENABLE_NETWORK_LEVEL_LOGGING = False
